#ifndef frame_buffer_object_h
#define frame_buffer_object_h

#include <GL/glew.h>
#include "object.h"

class FrameBufferObject : public Object
{
private:
  GLuint m_fbo;
  GLuint m_color;
  GLuint m_rb_depth;
  GLsizei m_width;
  GLsizei m_height;

  int m_viewport[4];
  int m_prev_target;


  void gen()
  {
    glGenTextures(1, &m_color);
    glBindTexture(GL_TEXTURE_2D, m_color);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, m_width, m_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glGenFramebuffersEXT(1, &m_fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_fbo);

    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_color, 0);

    glGenRenderbuffersEXT(1, &m_rb_depth);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_rb_depth);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, m_width, m_height);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, m_rb_depth);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
  }

public:
  FrameBufferObject(GLsizei width, GLsizei height) :
      m_fbo(0), m_color(0), m_rb_depth(0), m_width(width), m_height(height)
  {
    gen();
  }

  ~FrameBufferObject()
  {
    glDeleteRenderbuffersEXT(1, &m_rb_depth);
    glDeleteTextures(1, &m_color);
    glDeleteFramebuffersEXT(1, &m_fbo);
  }

  GLsizei getWidth() const { return m_width; }
  GLsizei getHeight() const { return m_height; }

  void enable()
  {
    // ݂viewportޔAFBÕTCYɍ킹 
    glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &m_prev_target);
    glGetIntegerv(GL_VIEWPORT, m_viewport);
    glViewport(0,0, getWidth(), getHeight());

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_fbo);
  }

  void disable()
  {
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_prev_target);

    // ޔviewport߂ 
    glViewport(m_viewport[0],m_viewport[1], m_viewport[2],m_viewport[3]);
  }

  void assign() { glBindTexture(GL_TEXTURE_2D, m_color); }
  void disassign() { glBindTexture(GL_TEXTURE_2D, 0); }
};


/*
 // gp: 

  FrameBufferObject *fbo;
  fbo = new FrameBufferObject(640, 480);


  fbo->enable();
  // FBOɕ`` 
  fbo->disable();
  fbo->assign(); // FBOeNX`ƂĎgp 
  glEnable(GL_TEXTURE_2D);
  // ` 
  glDisable(GL_TEXTURE_2D);


  // FBOeNX`ƂĎgpނƂł炵 
  fbo->enable();
  fbo->assign();
  glEnable(GL_TEXTURE_2D);
  // `揈 
  glDisable(GL_TEXTURE_2D);
  fbo->disable();
*/

#endif
