#ifndef SGUI_H
#define SGUI_H

#pragma warning(disable : 4996)




// FileDialogɑΉB
 #define SG_WITH_FILEDIALOG


#include <string>
#include <vector>
#include <list>
#include <map>

#include <SDL/SDL.h>
#include <FTGL/FTGLTextureFont.h>



namespace sgui {

typedef size_t wnd_id;
typedef FTFont Font;

using std::string;
using std::wstring;

struct Point
{
private:
  float m_x, m_y;

public:
  explicit Point(float x=0.0f, float y=0.0f) : m_x(x), m_y(y) {}
  float getX() const { return m_x; }
  float getY() const { return m_y; }
  void setX(float x) { m_x = x; }
  void setY(float y) { m_y = y; }

  void operator+=(const Point& p) { m_x+=p.m_x; m_y+=p.m_y; }
  void operator-=(const Point& p) { m_x-=p.m_x; m_y-=p.m_y; }
  Point operator+(const Point& p) const { Point r(*this); r+=p; return r; }
  Point operator-(const Point& p) const { Point r(*this); r-=p; return r; }
  Point operator*(float v) const { return Point(getX()*v, getY()*v); }
  Point operator/(float v) const { return Point(getX()/v, getY()/v); }
};

struct Size
{
private:
  float m_width, m_height;

public:
  explicit Size(float w=0.0f, float h=0.0f) : m_width(w), m_height(h) {}
  float getWidth() const { return m_width; }
  float getHeight() const { return m_height; }
  void setWidth(float w) { m_width = w; }
  void setHeight(float h) { m_height = h; }

  void operator+=(const Size& p) { m_width+=p.m_width; m_height+=p.m_height; }
  void operator-=(const Size& p) { m_width-=p.m_width; m_height-=p.m_height; }
  Size operator+(const Size& p) const { Size r(*this); r+=p; return r; }
  Size operator-(const Size& p) const { Size r(*this); r-=p; return r; }
  Size operator*(float v) const { return Size(getWidth()*v, getHeight()*v); }
  Size operator/(float v) const { return Size(getWidth()/v, getHeight()/v); }
};

struct Range
{
private:
  float m_min, m_max;

public:
  explicit Range(float n=0.0f, float x=0.0f) : m_min(n), m_max(x) {}
  float getMin() const { return m_min; }
  float getMax() const { return m_max; }
  float getLength() const { return getMax()-getMin(); }
  float clamp(float v) const { return std::max<float>(getMin(), std::min<float>(v, getMax())); }

  void setMin(float n) { m_min = n; }
  void setMax(float x) { m_max = x; }
};

struct Rect
{
private:
  float right, left, top, bottom;

public:
  explicit Rect(float r=0.0f, float l=0.0f, float t=0.0f, float b=0.0f) :
    right(r), left(l), top(t), bottom(b) {}
  explicit Rect(const Point& tl, const Point& br) :
    right(br.getX()), left(tl.getX()), top(tl.getY()), bottom(br.getY()) {}
  explicit Rect(const Point& pos, const Size& s) :
    right(pos.getX()+s.getWidth()), left(pos.getX()), top(pos.getY()), bottom(pos.getY()+s.getHeight()) {}
  explicit Rect(const Size& s) :
    right(s.getWidth()), left(0.0f), top(0.0f), bottom(s.getHeight()) {}

  float getLeft() const   { return left; }
  float getRight() const  { return right; }
  float getTop() const    { return top; }
  float getBottom() const { return bottom; }
  float getWidth() const  { return right-left; }
  float getHeight() const { return bottom-top; }
  Point getTopLeft() const { return Point(getLeft(), getTop()); }
  Point getTopRight() const { return Point(getRight(), getTop()); }
  Point getBottomLeft() const { return Point(getLeft(), getBottom()); }
  Point getBottomRight() const { return Point(getRight(), getBottom()); }
  Size getSize() const { return Size(getWidth(), getHeight()); }

  bool isInside(const Point& p) const
  {
    return (p.getX()>=getLeft() && p.getX()<=getRight() && p.getY()>=getTop() && p.getY()<=getBottom());
  }

  bool isInside(const Rect& r) const
  {
    return (isInside(r.getTopLeft())
      && isInside(r.getTopRight())
      && isInside(r.getBottomLeft())
      && isInside(r.getBottomRight()));
  }

  bool isOverlaped(const Rect& r) const
  {
    return (getRight()>=r.getLeft()
      && getLeft()<=r.getRight()
      && getTop()<=r.getBottom()
      && getBottom()>=r.getTop());
  }

  void clip(Rect& r)
  {
    if(!isOverlaped(r)) {
      r = Rect();
      return;
    }

    if(r.left<left)  r.left = left;
    if(r.right>right) r.right = right;
    if(r.top<top)      r.top = top;
    if(r.bottom>bottom) r.bottom = bottom;
  }

  void operator+=(const Point& p) { right+=p.getX(); left+=p.getX(); top+=p.getY(); bottom+=p.getY(); }
  void operator-=(const Point& p) { right-=p.getX(); left-=p.getX(); top-=p.getY(); bottom-=p.getY(); }
  Rect operator+(const Point& p) const { Rect r(*this); r+=p; return r; }
  Rect operator-(const Point& p) const { Rect r(*this); r-=p; return r; }
};

struct Color
{
public:
  union {
    struct { float r, g, b, a; };
    float v[4];
  };

  explicit Color(float _r=1.0f, float _g=1.0f, float _b=1.0f, float _a=1.0f)
  {
    r=_r; g=_g; b=_b; a=_a;
  }
};




class Mutex
{
private:
  SDL_mutex *m_mutex;

  const Mutex& operator=(const Mutex&); // Rs[֎~ 
  Mutex(const Mutex&);

public:
  Mutex() { m_mutex = SDL_CreateMutex(); }
  ~Mutex() { SDL_DestroyMutex(m_mutex); }
  void lock() { SDL_LockMutex(m_mutex); }
  void unlock() { SDL_UnlockMutex(m_mutex); }
  SDL_mutex* get() { return m_mutex; }
};

class ScopedLock
{
private:
  Mutex& m_mutex;

public:
  ScopedLock(Mutex& m) : m_mutex(m) { m_mutex.lock(); }
  ~ScopedLock() { m_mutex.unlock(); }
  Mutex& get() { return m_mutex; }
};



void AssignColor(const Color& c);
void AssignColorO(const Color& c, float opa);
void DrawPoint(const Point& p);
void DrawLine(const Point& p1, const Point& p2);
void DrawRectEdge(const Rect& r);
void DrawRect(const Rect& r);







//  RootObject

class Object
{
public:
  Object();
  virtual ~Object();
  virtual void release();
};


//  Event

class EventServer;
class EventClient;
class Window;
class ListItem;
class DDItem;

enum EventType {
  EVT_UNKNOWN = 0,

  EVT_CONSTRUCT,
  EVT_DESTRUCT,

  EVT_APP_DESTROY_WINDOW,
  EVT_APP_EXIT,
  EVT_APP_RESIZE,
  EVT_GAININGKEYFOCUS,
  EVT_GAININGMOUSEFOCUS,
  EVT_LOSINGKEYFOCUS,
  EVT_LOSINGMOUSEFOCUS,

  EVT_MOUSE_BUTTONDOWN,   // MouseEvent
  EVT_MOUSE_BUTTONUP,     // MouseEvent
  EVT_MOUSE_MOVE,         // MouseEvent
  EVT_MOUSE_DOUBLECLICKED,// MouseEvent
  EVT_KEYDOWN,            // KeyboardEvent
  EVT_KEYUP,              // KeyboardEvent

  EVT_JOY_BUTTONDOWN,     // JoyEvent
  EVT_JOY_BUTTONUP,       // JoyEvent
  EVT_JOY_AXIS,           // JoyEvent

  EVT_TIMER,              // TimerEvent

  EVT_RESIZE,             // SizeEvent
  EVT_BUTTON_DOWN,        // Event
  EVT_BUTTON_UP,          // Event
//  EVT_CHECK,            // Event
//  EVT_CHECK_UNCHECK,    // Event
  EVT_EDIT,               // EditEvent
  EVT_EDIT_ENTER,         // EditEvent
  EVT_SCROLL,             // ScrollEvent
//  EVT_SLIDER,           // ScrollEvent
//  EVT_RADIOBUTTON,      // Event
  EVT_LIST_SELECT,        // ListEvent
  EVT_LIST_DISSELECT,     // ListEvent
  EVT_LIST_DOUBLECLICK,   // ListEvent
  EVT_COMBO,              // ListEvent
  EVT_DD_DRAG,            // DragAndDropEvent
  EVT_DD_DRAGING,         // DragAndDropEvent
  EVT_DD_DROP,            // DragAndDropEvent
  EVT_DD_RECIEVE,         // DragAndDropEvent
  EVT_CLOSEDIALOG,        // Event
  EVT_MESSAGEDIALOG,      // DialogEvent
  EVT_CONFIRMDIALOG,      // DialogEvent
  EVT_FILEDIALOG,         // DialogEvent

  EVT_USER, // UserEvent
};

class Event : public Object
{
friend class EventServer;
private:
  const EventType m_type;
  Window *m_src;
  Window *m_dst;
  int m_hcontrol;
  int m_hframe;

public:
  Event(EventType type, Window *src, Window *dst);

  EventType getType() const;
  Window* getSrc() const;
  Window* getDst() const;
  int isHandled() const;
  int isHandledByControl() const;
  int isHandledByFrame() const;
};

class DestructEvent : public Event
{
friend class EventServer;
private:
  Window *m_self;
  int m_selfid;

public:
  DestructEvent(EventType type, Window *src, Window *dst, Window *self);

  // deleteς݂̃IuWFNgւ̃|C^ԂB|C^l̔rȊOɎg_ 
  Window* getSelf() const { return m_self; }
  int getSelfID() { return m_selfid; }
};

class KeyboardEvent : public Event
{
private:
  unsigned char m_scan_code;
  SDLMod m_modifiers;
  SDLKey m_key;
  Uint16 m_unicode;

public:
  KeyboardEvent(EventType MessageType, Window *src, Window *dst,
    unsigned char scan_code, SDLMod modifiers, SDLKey key, Uint16 unicode);

  unsigned char getCode() const { return m_scan_code; }
  SDLMod getModifier() const { return m_modifiers; }
  SDLKey getKey() const { return m_key; }
  Uint16 getUnicode() const { return m_unicode; }
};

class MouseEvent : public Event
{
private:
  Point m_position;
  Point m_relative;
  Point m_real_position;
  Point m_real_relative;
  unsigned int m_button;

public:
  enum Button
  {
    NONE     = 0,
    LEFT     = 1<<0,
    MIDDLE   = 1<<1,
    RIGHT    = 1<<2,
    WHEELUP  = 1<<3,
    WHEELDOWN= 1<<4,
  };

  MouseEvent(EventType type, Window *src, Window *dst, Point pos, Point relative, unsigned int button);
  static unsigned int TranslateSDLButton(Uint8 sdl_button);
  static unsigned int TranslateSDLButtonState(Uint8 sdl_button_state);

  const Point& getPosition() const { return m_position; } // OpenGLW 
  const Point& getRelative() const { return m_relative; } //
  const Point& getRealPosition() const { return m_real_position; } // WindowW 
  const Point& getRealRelative() const { return m_real_relative; } //
  unsigned int getButton() const { return m_button; }
};

class JoyEvent : public Event
{
private:
  int m_devid;
  int m_button;
  int m_axis_x;
  int m_axis_y;

public:
  JoyEvent(EventType type, Window *src, Window *dst, int device, int button, int x, int y);
  int getButton() const { return m_button; }
  int getAxisX() const { return m_axis_x; }
  int getAxisY() const { return m_axis_y; }
};



class DragAndDropEvent : public MouseEvent
{
private:
  DDItem *m_item;

public:
  DragAndDropEvent(EventType type, Window *src, Window *dst, Point pos, Point relative, unsigned int button, DDItem *item) :
    MouseEvent(type, src, dst, pos, relative, button), m_item(item)
  {}

  DDItem* getItem() const { return m_item; }
};

class ResizeEvent : public Event
{
private:
  Size m_size;
  Size m_relative;

public:
  ResizeEvent(EventType type, Window *src, Window *dst, Size size, Size rel) :
    Event(type, src, dst), m_size(size), m_relative(rel)
  {}
  const Size& getSize() const { return m_size; }
  const Size& getRelative() const { return m_relative; }
};

class TimerEvent : public Event
{
public:
  TimerEvent(EventType type, Window *src, Window *dst) :
    Event(type, src, dst)
  {}
};

class EditEvent : public Event
{
private:
  wstring m_string;

public:
  EditEvent(EventType type, Window *src, Window *dst, const wstring& str) :
    Event(type, src, dst), m_string(str)
  {}
  const wstring& getString() const { return m_string; }
};

class ScrollEvent : public Event
{
public:
  ScrollEvent(EventType type, Window *src, Window *dst, float value, float rel) :
    Event(type, src, dst), m_value(value), m_relative(rel)
  {}
  float getValue() const { return m_value; }
  float getRelative() const { return m_relative; }

private:
  float m_value;
  float m_relative;
};

class ListEvent : public Event
{
public:
  ListEvent(EventType type, Window *src, Window *dst, size_t index, ListItem *item) :
    Event(type, src, dst), m_index(index), m_item(item)
  {}

  size_t getIndex() const { return m_index; }
  ListItem* getItem() const { return m_item; }

private:
  size_t m_index;
  ListItem *m_item;
};

class DialogEvent : public Event
{
public:
  DialogEvent(EventType type, Window *src, Window *dst, bool ok, const string& path="") :
    Event(type, src, dst), m_ok(ok), m_path(path)
  {}

  bool isOK() { return m_ok; }
  const string& getPath() const { return m_path; }

private:
  bool m_ok;
  string m_path;
};

class UserEvent : public Event
{
public:
  UserEvent(EventType type, Window *src, Window *dst, const string& name, void *value) :
    Event(type, src, dst), m_name(name), m_value(value)
  {}
  const string& getName() const { return m_name; }
  void* getValue() const { return m_value; }

private:
  string m_name;
  void *m_value;
};













//  EventServer

class EventServer
{
private:
  typedef std::list<EventClient*> eventclient_cont;
  typedef std::map<EventType, eventclient_cont> eventclient_map;
  typedef std::list<Event*> event_cont;

  eventclient_map m_evtclients;
  eventclient_cont m_toplevel;
  event_cont m_events;
  Mutex m_mutex;

public:
  EventServer();
  ~EventServer();
  void registerEventClient(EventClient *client, EventType type);
  void deregisterEventClient(EventClient *client, EventType type);
  void deregisterEventClient(EventClient *client);
  void setTopLevel(EventClient *client);

  void queueEvent(Event *evt); // synchronized
  void deliverEvent(); // synchronized
  void clearEvent(); // synchronized
};











class FontRenderer
{
public:
  enum HAlign {
    HLEFT,
    HCENTER,
    HRIGHT,
  };
  enum VAlign {
    VTOP,
    VCENTER,
    VBOTTOM,
  };

  FontRenderer();
  void setHAlign(HAlign align);
  void setVAlign(VAlign align);
  void setFont(Font *font);
  void draw(const wstring& str, const Point& pos=Point(), const Rect& edge=Rect()) const;

  Font* getFont() const;
  float getHeight() const;
  float getHeight(const wstring& str) const;
  float getAdvance(wchar_t c) const;
  float getAdvance(const wchar_t *str, size_t len) const;
  float getAdvance(const wstring& str) const;

  Font* m_font;
  HAlign m_halign;
  VAlign m_valign;
};

class StopWatch
{
public:
  StopWatch();
  void run(size_t limit);
  void reset();
  void stop();
  bool isRunning() const;
  bool isFinished() const;
  bool isActive() const;
  size_t getLeft() const;
  size_t getPast() const;

private:
  size_t m_start_time;
  size_t m_end_time;
};

class Timer
{
public:
  static Uint32 callback(Uint32 interval, void *param);

  Timer();
  ~Timer();
  void run(size_t interval, bool autorestart=false);
  void stop();
  void onTimer(size_t interval);

  void setTarget(Window *target);

private:
  SDL_TimerID m_timerid;
  Window *m_target;
  bool m_autorestart;
};








class EventClient : public Object
{
public:
  EventClient();
  virtual ~EventClient();
  virtual void listenEvent(EventType type);
  virtual void queueEvent(Event *e);
  virtual bool handleEvent(const Event& evt);
};



//  Window

class Window : public EventClient
{
public:
  typedef std::list<Window*> window_cont;

  Window(Window *parent, const Point& pos=Point(), const Size& size=Size(), const wstring& text=L"", wnd_id id=0);
  virtual ~Window();
  virtual void releaseChildren();
  virtual void update();
  virtual void draw();
  virtual float getDrawPriority() const;
  virtual void destroy();

  virtual bool isInside(const Point& global) const;
  virtual Point toGlobalCoord(const Point& local) const;
  virtual Point toLocalCoord(const Point& global) const;
  virtual void clip(Rect& r, const Window *wnd=0) const;
  virtual void fitViewport() const;

  virtual wnd_id getID() const;
  virtual const Color&   getForegroundColor() const;
  virtual const Color&   getBackgroundColor() const;
  virtual const Color&   getBorderColor() const;
  virtual const wstring& getText() const;
  virtual const Point&   getPosition() const;
  virtual const Size&    getSize() const;
  virtual const Point&   getClientPosition() const;
  virtual const Size&    getClientSize() const;
  virtual bool isVisible() const;
  virtual bool isFloating() const;
  virtual Point getGlobalPosition() const;

  virtual void setForegroundColor(const Color& color);
  virtual void setBackgroundColor(const Color& color);
  virtual void setBorderColor(const Color& color);
  virtual void setText(const wstring& text);
  virtual void setPosition(const Point& pos);
  virtual void setSize(const Size& size);
  virtual void setFloating(bool floating);
  virtual void setVisible(bool visible);
  virtual void setTopLevel(Window *window);
  virtual void move(const Point& rel);
  virtual void toTopLevel();
  virtual void toTopLevelEvent();

  virtual Window* getParent() const;
  virtual window_cont& getChildren();
  virtual Window* getChildByID(wnd_id id) const;
  virtual void setParent(Window *parent);
  virtual void registerChild(Window *child);
  virtual void deregisterChild(Window *child);

  virtual Font* getFont() const;
  virtual void setFont(Font* font);

protected:
  FontRenderer m_fontr;
  wnd_id m_id;
  wstring m_text;
  Color m_fg_color;
  Color m_bg_color;
  Color m_bd_color;
  Point m_position;
  Size m_size;
  bool m_visible;
  bool m_floating;

  Window *m_parent;
  window_cont m_children;
};








//  Label

class Label : public Window
{
public:
  Label(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id=0);
  virtual void draw();
};




//  Button
//
//  EVT_BUTTON_DOWN: {^ꂽƂ 
//  EVT_BUTTON_UP: {^ꂽƂ 

class Button : public Window
{
public:
  enum State {
    UP,
    DOWN,
    DISABLED,
  };

  Button(Window *parent, const Point& pos, const Size& size, const wstring& text=L"", wnd_id id=0);
  virtual void draw();
  virtual bool handleEvent(const Event& evt);

  virtual void up();
  virtual void down();
  virtual State getButtonState() const;
  virtual void setButtonState(State state);

private:
  State m_button_state;
};


class ToggleButton : public Button
{
public:
  ToggleButton(Window *parent, const Point& pos, const Size& size, const wstring& text=L"", wnd_id id=0);
  virtual bool handleEvent(const Event& evt);

  virtual void toggle();

private:
  bool m_clicked;
};




//  RadioButton

/*
enum {
  RB_GROUP = 1,
};

class RadioButton : public Window
{
public:
  RadioButton(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id=0, int style=0);
  virtual void draw();
  virtual bool handleEvent(const Event& evt);

private:
};
*/


//  ToolTip

class ToolTip : public Window
{
public:
  ToolTip(Window *parent, const wstring& text, wnd_id id=0);
  virtual bool handleEvent(const Event& evt);
  virtual void draw();
  virtual float getDrawPriority() const;

private:
  Timer m_timer;
  size_t m_appear_time;
};




//  Edit
//
//  EVT_EDIT : ҏWꂽƂ 
//  EVT_ENTER : tH[JXĂEnterꂽƂ 

class Edit : public Window
{
public:
  Edit(Window *parent, const Point& pos, const Size& size, const wstring& text=L"", wnd_id id=0);
  virtual void draw();
  virtual bool handleEvent(const Event& evt);

  virtual void focus();
  virtual void defocus();

  virtual bool isReadOnly() const;
  virtual bool isFocused() const;
  virtual void setReadOnly(bool ro);
  virtual void setBlinkerPos(int pos);
  virtual void setText(const wstring& text);

private:
  bool m_readonly;
  int m_blinker;
  bool m_focused;
};




//  DDItem
//
//  EVT_DD_DRAG : NbNꂽƂ 
//  EVT_DD_DRAGING : hbOꂽƂ 
//  EVT_DD_DROP : hbvꂽƂ 

class DDItem : public Window
{
public:
  DDItem(Window *parent, const Point& pos, const Size& size, const wstring& text=L"", wnd_id id=0);
  virtual void draw();
  virtual bool handleEvent(const Event& evt);
  virtual bool isDragging() const;
  virtual const Point& getDragBeginPosition() const;

private:
  StopWatch m_stopwatch; // _uNbNop 
  Point m_drag_pos;
  bool m_dragging;
};



//  DDReciever
//
//  EVT_DD_RECIEVE : DDItem͈͓ɃhbvꂽƂ 

class DDReciever : public Window
{
public:
  DDReciever(Window *parent, const Point& pos, const Size& size, const wstring& text=L"", wnd_id id=0);
  virtual void draw();
  virtual bool handleEvent(const Event& evt);
  virtual void deregisterChild(Window *child);

  virtual void setItem(DDItem *item);
  virtual DDItem* getItem() const;

private:
  DDItem *m_item;
};



//  ScrollBar
//
//  EVT_SCROLL : l(setValue)ől(setRange)ύXꂽƂ 

enum {
  SB_HORIZONTAL = 0,
  SB_VERTICAL = 1,
};

class ScrollBar : public Window
{
public:
  enum {
    HORIZONTAL,
    VERTICAL,
  };

  ScrollBar(Window *parent, int style=0, float size=0.0f, wnd_id id=0);
  virtual void draw();
  virtual bool handleEvent(const Event& event);

  virtual void fitTo(Window *target);
  virtual float getValue() const;
  virtual float getThick() const;
  virtual float getStep() const;
  virtual float getRange() const;

  virtual void setScrollBarSize(float size);
  virtual void setThick(float value);
  virtual void setValue(float value);
  virtual void setStep(float value);
  virtual void setRange(float range);

protected:
  float getBarWidth() const;
  Range getScrollBarRange() const;
  void resetBarRect();
  void mouseToScroll(const Point& mouse_move);

private:
  int m_align;
  bool m_dragging;
  float m_thick;
  float m_step;
  float m_scroll_pos;
  float m_range;
  float m_bar_width;
  Button *m_button_tl;
  Button *m_button_br;
  Rect m_bar;
};




enum {
  VSCROLLBAR = 1,
  HSCROLLBAR = 2,
};

class ScrolledWindow : public Window
{
public:
  ScrolledWindow(Window *parent, const Point& pos, const Size& size,
    const wstring& text=L"", wnd_id id=0, int style=VSCROLLBAR);
  virtual void draw();
  virtual bool handleEvent(const Event& event);

  virtual const Point& getBasePosition() const;
  virtual const Size& getSizeNS() const; // XN[o[̕Size 
  virtual const Size& getScrollSize() const;
  virtual ScrollBar* getVScrollBar() const;
  virtual ScrollBar* getHScrollBar() const;
  virtual void clip(Rect& r, const Window *wnd) const;

  virtual void setSize(const Size& size);
  virtual void setScrollSize(const Size& size);
  virtual void scroll(const Point& rel);

private:
  Point m_base_pos;
  Size m_scroll_size;
  Size m_size_ns;
  ScrollBar *m_hscrollbar;
  ScrollBar *m_vscrollbar;
};




class ListItem : public Window
{
public:
  ListItem(const wstring& text);
  virtual void draw();

  virtual bool isSelected() const;
  virtual void setSelect(bool selected);

  bool m_selected;
};


//  List
//
//  EVT_LIST_SELECT : vfIꂽƂ 
//  EVT_LIST_DISSELECT : vfIԂɂȂƂ 
//  EVT_LIST_DOUBLECLICK : vf_uNbNꂽƂ 

enum {
  LS_MULTISELECT = 1,
};

class List : public ScrolledWindow
{
public:
  typedef std::vector<ListItem*> item_cont;

  List(Window *parent, const Point& pos, const Size& size, wnd_id id=0, int style=0);
  virtual void draw();
  virtual bool handleEvent(const Event& event);

  virtual float getItemHeight() const;
  virtual ListItem* getItem(size_t i);
  virtual ListItem* getSelectedItem(size_t i=0);
  virtual size_t getItemCount() const;
  virtual size_t getItemIndex(ListItem *l) const;
  virtual size_t getSelection(size_t i=0) const;
  virtual size_t getSelectionCount() const;
  virtual size_t hitItem(const Point& p) const;

  virtual void addItem(ListItem *item);
  virtual void removeItem(size_t i);
  virtual void removeItem(ListItem *i);
  virtual void selectItem(size_t i);
  virtual void selectItem(ListItem *i);
  virtual void clearItem();
  virtual void clearSelection();
  virtual void alignItems();
  virtual void resetScrollSize();

private:
  bool m_multiselect;
  float m_item_height;
  item_cont m_items;

  StopWatch m_stopwatch; // _uNbNop 
  ListItem *m_last_selected;
};



//  Combo
//
//  EVT_COMBO : vfIƂ 

class Combo : public Window
{
public:
  Combo(Window *parent, const Point& pos, const Size& size, wnd_id id=0, int style=0);
  virtual bool handleEvent(const Event& event);
  virtual void draw();

  virtual void addItem(ListItem *item);
  virtual void clearItem();

  virtual List* getList();
  virtual ListItem* getSelection();
  virtual void setSelection(size_t i);

private:
  Button *m_list_button;
  List *m_list;
  ListItem *m_selection;
};



//  Panel
//
//  EVT_RESIZE : TCYύXꂽƂ 

class Panel : public Window
{
public:
  Panel(Window *parent, const Point& pos, const Size& size,
    const wstring& text=wstring(), wnd_id id=0, int style=0);
  virtual void draw();
  virtual bool handleEvent(const Event& event);
  virtual void setSize(const Size& size);
};



//  Dialog
//
//  EVT_CLOSEDIALOG : Ƃ 

enum {
  MOVABLE = 1,
  RESIZABLE = 2,
  CLOSE_BUTTON = 4,
  MAXIMIZE_BUTTON = 8,
  MINIMIZE_BUTTON = 16,
  STAY_ON_TOP = 32,
};

class Dialog : public Panel
{
public:
  Dialog(Window *parent, const Point& pos, const Size& size,
    const wstring& text=L"", wnd_id id=0, int style= MOVABLE | RESIZABLE | CLOSE_BUTTON);
  virtual void destroy();
  virtual void draw();
  virtual bool handleEvent(const Event& event);

  virtual const Point& getClientPosition() const;
  virtual const Size& getClientSize() const;
  virtual void setSize(const Size& size);
  virtual void setWidthRange(const Range& r);
  virtual void setHeightRange(const Range& r);

  virtual void onCloseButton();
  virtual void onMaximizeButton();
  virtual void onMinimizeButton();

protected:
  Point m_client_pos;
  Size m_client_size;
  Button *m_close_button;
  Button *m_maximize_button;
  Button *m_minimize_button;

  bool m_movable;
  bool m_moving;

  bool m_resizable;
  int m_resizing;
  Range m_width_range;
  Range m_height_range;
};

//  MessageDialog

class MessageDialog : public Dialog
{
public:
  MessageDialog(const wstring& title, const wstring& message, Window *parent=0,
    const Point& pos=Point(), const Size& size=Size(320, 200));
  virtual bool handleEvent(const Event& event);

protected:
  Button *m_ok_button;
  Label *m_message;
};


//  ConfirmDialog
//
//  EVT_CONFIRMDIALOG

class ConfirmDialog : public Dialog
{
public:
  ConfirmDialog(const wstring& title, const wstring& message, Window *parent=0,
    const Point& pos=Point(), const Size& size=Size());
  virtual bool handleEvent(const Event& event);

protected:
  Button *m_ok_button;
  Button *m_cancel_button;
  Label *m_message;
};


//  FileDialog
//
//  EVT_FILEDIALOG
#ifdef SG_WITH_FILEDIALOG

bool IsFile(const string& path);
bool IsDir(const string& path);
bool MakeDir(const string& path);
bool MakeDeepDir(const string& path);
bool RemoveDir(const string& path);
bool Remove(const string& path);
string GetCWD();

wstring _L(const string& src);
string _S(const wstring& src);

class Dir
{
public:
  typedef std::vector<string> path_cont;
  typedef path_cont::iterator iterator;
  typedef path_cont::const_iterator const_iterator;

  Dir();
  explicit Dir(const string& path);
  bool open(const string& path);
  bool openRecursive(const string& path);

  size_t size() const;
  const string& operator[](size_t i) const;
  const string& getPath() const;

  iterator begin() { return m_files.begin(); }
  iterator end()   { return m_files.end(); }
  const_iterator begin() const { return m_files.begin(); }
  const_iterator end() const   { return m_files.end(); }

private:
  string m_path;
  path_cont m_files;
};


class FileDialog : public Dialog
{
public:
  FileDialog(const wstring& title, Window *parent=0, const Point& pos=Point(), const Size& size=Size());
  virtual bool handleEvent(const Event& event);
  virtual bool openDir(const string& path);

protected:
  Button *m_ok_button;
  Button *m_cancel_button;
  List *m_filelist;
  Combo *m_locate;
  string m_path;
  string m_file;
};

#endif


//  View

class View : public Window
{
public:
  static View* instance();

  View(const string& title, const Size& window_size, const Size& gl_size=Size(), bool full=false);
  virtual void draw();

  virtual void setViewport(const Point& pos, const Size& size) const;
  virtual Point toOpenGLCoord(const Point& point) const; // WindowWOpenGLWɕϊ 
  virtual Point toWindowCoord(const Point& point) const; // OpenGLWWindowWɕϊ 
  virtual void setSize(const Size& size);

  const Size& getWindowSize() const;

private:
  static View *m_instance;
  float m_width_scale;
  float m_height_scale;
  bool m_fullscreen;
  Size m_window_size;
};



/*
  App
*/

class App : public Object
{
public:
  static App* instance();

  App(int argc=0, char **argv=0);
  virtual ~App();

  virtual void loop();
  virtual void update();
  virtual void draw();

  virtual bool handleEvent(const Event& event);
  virtual int exec();
  virtual void stop();
  virtual void initOpenGLResource();

  virtual EventServer& getEventServer();
  virtual const Color& getDefaultForegroundColor() const;
  virtual const Color& getDefaultBackgroundColor() const;
  virtual const Color& getDefaultBorderColor() const;
  virtual Font* getDefaultFont() const;
  virtual Font* getFont(const string& path, size_t size) const;
  virtual Window* getKeyFocus() const;
  virtual Window* getMouseFocus() const;
  virtual View* getView();

  virtual bool loadFont(const string& path, size_t size);
  virtual void releaseFont(const string& path);
  virtual void setDefaultForegroundColor(const Color& color);
  virtual void setDefaultBackgroundColor(const Color& color);
  virtual void setDefaultBorderColor(const Color& color);
  virtual void setDefaultFont(Font *font);
  virtual void setDefaultFont(const string& name);
  virtual void setDefaultFontSize(size_t);
  virtual void setKeyFocus(Window *window);
  virtual void setMouseFocus(Window *window);
  virtual void setView(View *view);


protected:
  typedef std::map<string, Font*> font_cont;

  void waitSDLEvent();

  static App *m_instance;

  EventServer m_evtserver;
  int m_argc;
  char **m_argv;

  Color m_fg_color;
  Color m_bg_color;
  Color m_bd_color;
  string m_default_font_name;
  size_t m_default_font_size;
  Font *m_default_font;
  font_cont m_fonts;

  bool m_end_flag;
  View *m_view;
  Window *m_key_focus_window;
  Window *m_mouse_focus_window;
};


} // sgui



#endif // SGUI_H
